import csv
from datetime import datetime, timedelta

import matplotlib.pyplot as plt
from tabulate import tabulate
import platform


def configuration():
    """
    Config start
    Endre konfigurasjon for å endre visualisering
    """

    # Filnavn. Legg til hele filbanen hvis filen ligger en annen plass.
    filename = "Log_22_orginal.csv"

    # Datoer
    # Skal filen filtreres med de satte dato-instillingene
    date_filter = False
    # For hvilken kolonne i filen ligger datoen.
    date_column = 1
    # Start dato
    start_date = "01.01.2022"
    # Slutt dato
    end_date = "31.12.2022"

    # Velger hvilke tidsrom. Fra 00:00:00 til 00:00:00 er hele døgnet.
    start_time = "07:30:00"
    end_time = "11:00:00"
    # Velger hvilke ukedager
    days = [0, 2, 4]

    # Inverterer datofilteret. Hvis det er filtrert for alle hverdager i arbeidstid vil man ved å invertere få alle
    # hverdager utenfor arbeidstid og hele helga. Standard = True
    inside_time_range = False
    # Vis kalender før resultat for å dobbelsjekke at korrekt dato er oppgitt
    show_calendar = True

    # Merk: Svarteliste kjøres før hviteliste. Plukker filteret opp et svartelistet ord vil raden fjernes uansett hva
    # som måtte stå i hvitelisten.

    # Velger om filtreringen skal ta hensyn til små og store bokstaver
    case_sensitive = False
    # Velg hvilke ord som skal svartelistes
    blacklist = ["Warning", "Off", "acknowledge"]
    # Velg hvilke ord som skal hvitelistes
    whitelist_any = []

    # Velger hvilken kombinajson av ord som skal hvitelistes. Setningne må inneholde samtlige ord.
    whitelist_all = []

    # Antall timer det må gå før samme alarm telles igjen, desimaltall fungerer (1/60 for 1 minutt).
    # Ønsker du alle treff, sett hours_between_alarms = 0
    hours_between_alarms = 0.5

    # Hvilke kolonner skal vises og sorteres i søylediagrammet
    show_columns = [5]

    # Velge antall søyler som skal vises i søylediagrammet
    bar_count = 20

    # Nøkkelord som skal farges, nøkkelord / nøkkekombinasjoner i samme liste vil få samme farge.
    keywords = ["Nivå", ["%SAT", "O2"], "Temp", "pH", ["uv", "UV", "Trykk", "TRYKK"], ["Flow", "FLOW"]]
    # Tar kolonnenavnene fra første rad i csv-filen og bruker de til y-aksenavn
    auto_label = True
    # Hvis auto label er av, sett eget aksenavn
    custom_label = ""

    # Farger på søylene.
    colors = ["royalblue", "deepskyblue", "maroon", "forestgreen", "goldenrod", "darkorange"]

    """
    Config slutt
    Konfigurasjonsinnstillengene er slutt, det er ikke meningen å endre kode under dette punktet.
    """

    if date_filter and show_calendar:
        create_calendar(start_time, end_time, days, inside_time_range)
    # Funksjon som filtrerer data.
    label, data = filter_file(filename, auto_label, blacklist, whitelist_any, whitelist_all, date_filter, start_date,
                              end_date, start_time, end_time, days, inside_time_range, show_columns,
                              hours_between_alarms, case_sensitive, date_column)

    print("Summen av antall treff:", sum(data.values()))
    if len(custom_label) > 0:
        label = custom_label

    # Gjenta fargevalgene for nøkkelord for å sikre at det er nok farger tilgjengelig
    colors = colors * (len(keywords) // len(colors) + 1)
    colors = colors[:len(keywords)]

    day_names = ['Man.', 'Tir.', 'Ons.', 'Tor.', 'Fre.', 'Lør.', 'Søn.']
    day_list = "".join(day_names[day] for day in days)
    if inside_time_range:
        if len(days) == 7:
            time_title = f"Alle dager fra {start_time} til {end_time}"
        else:
            time_title = f"{day_list}, fra {start_time} til {end_time}"
    else:
        if len(days) == 7:
            time_title = f"Alle dager fra {end_time} til {start_time}"
        else:
            time_title = f"Alle tider, ekskludert {day_list}, fra {start_time} til {end_time}"

    time_title = time_title + len(time_title) * " "

    # Genererer søylediagrammet basert på iltrerte data
    draw_bar_chart(data, date_filter, bar_count, show_columns, label, keywords, colors, time_title)


def filter_row(row, blacklist, whitelist_any, whitelist_all, case_sensitive):
    """
    Sjekker om rader innholder elementer som ikke skal være med i resultatet.
    :param row: Rad innlest fra CSV-filen
    :param blacklist: Liste med svartelistede elementer
    :param whitelist_any: Liste med hvitelistede elementer
    :param whitelist_all: Liste der alle elementer må være til stede i raden for å hviteliste raden
    :param case_sensitive: Boolsk parameter som bestemmer om filteret skal være case-sensitiv.
    :return: Returnerer True om raden skal filtreres bort.
    """

    # Lager en kopi av raden for å unngå å endre originalen
    row = row.copy()

    # Gjør filteret case-sensitivt hvis nødvendig
    if not case_sensitive:
        blacklist = [word.lower() for word in blacklist]
        whitelist_any = [word.lower() for word in whitelist_any]
        whitelist_all = [word.lower() for word in whitelist_all]
        row[0] = row[0].lower()

    # Sjekker om raden inneholder svartelistede elementer
    for word in blacklist:
        if word in row[0]:
            return True

    # Sjekker om raden inneholder hvitelistede elemeneter
    for word in whitelist_any:
        if word in row[0]:
            return False

    # Sjekker om raden inneholder alle elementene i hvitelisten
    for word in whitelist_all:
        if word not in row[0]:
            return True

    # Hvis hvitelisten er tom, blir raden filtrert bort
    if len(whitelist_all) > 0:
        return False

    # Hvis hvitelisten er tom, blir raden beholdt hvis det ikke finnes noen elementer i whitelist_any
    if len(whitelist_any) > 0:
        return True
    else:
        return False


def filter_date(row, start_date_str, end_date_str, start_time_str, end_time_str, days, inside_time_range,
                date_column):
    """
    Sjekker om raden oppfyller kriteriene for filtrering basert på dato og tid.
    :param row: Rad innlest fra CSV-filen.
    :param start_date_str: Startdato i strengformat (dd.mm.åååå).
    :param end_date_str: Sluttdato i strengformat (dd.mm.åååå).
    :param start_time_str: Starttidspunkt i strengformat (hh:mm:ss).
    :param end_time_str: Slutttidspunkt i strengformat (hh:mm:ss).
    :param days: Liste med ukedager (0-6, der 0 er mandag og 6 er søndag).
    :param inside_time_range: Boolsk parameter som bestemmer om tidspunktet må være innenfor eller utenfor tidsspekteret.
    :param date_column: Kolonneindeks der datoen finnes i raden.
    :return: Returnerer True hvis raden oppfyller filtreringskriteriene, ellers False.
    """

    # Legger til klokkeslett der dette mangler
    date_str = row[0].split(";")[date_column]
    if " " not in date_str:
        date_str += " 00:00:00"
    if len(date_str) == 16:
        date_str += ":00"

    # Konverter datoer og tidspunkter til datetime-objekter
    date_obj = datetime.strptime(date_str, "%d.%m.%Y %H:%M:%S")
    start_datetime_obj = datetime.strptime(start_date_str + " 00:00:00", "%d.%m.%Y %H:%M:%S")
    end_datetime_obj = datetime.strptime(end_date_str + " 23:59:59", "%d.%m.%Y %H:%M:%S")
    start_time_obj = datetime.strptime(start_time_str, "%H:%M:%S").time()
    end_time_obj = datetime.strptime(end_time_str, "%H:%M:%S").time()

    # Hvis sluttdatoen er før startdatoen, er parameterne ugyldige
    if end_datetime_obj <= start_datetime_obj:
        raise ValueError("Sluttdatoen må være etter startdatoen.")

    # Hvis nåværende tidspunkt ikke er innenfor start- og sluttdatoen, returner False
    if not start_datetime_obj <= date_obj <= end_datetime_obj:
        return False

    day_bool = False
    time_bool = False
    if date_obj.weekday() in days:
        day_bool = True

    # Hvis slutttidspunktet er før starttidspunktet, er vi over midnatt
    if end_time_obj <= start_time_obj:
        # Sjekker om tidspunktet er etter starttidspunktet og før sluttidspunktet
        if date_obj.time() >= start_time_obj or date_obj.time() <= end_time_obj:
            time_bool = True
    # Sjekker om tidspunktet er mellom start- og sluttidspunktet
    elif start_time_obj <= date_obj.time() <= end_time_obj:
        time_bool = True

    if inside_time_range:
        # Hvis tidspunktet er innenfor tidsspekteret og riktig ukedag
        if day_bool and time_bool:
            return True
        else:
            return False
    else:
        # Hvis tidspunktet er utenfor tidsspekteret eller feil ukedag
        if day_bool and time_bool:
            return False
        else:
            return True


def filter_duplicates(data, hours_between_alarms, date_column):
    """
    Fjerner duplikater fra data basert på tidsintervall mellom alarmer
    :param data: Data etter filtrering
    :param hours_between_alarms: Tiden det må gå mellom to alarmer for at den skal telles med, i timer.
    :param date_column: Kolonneindeks der datoen finnes i raden
    :return: Filtrerte data uten duplikater
    """

    last_seen_data = {}  # Holder oversikt over sist sett data for hver meta-informasjon
    result = []  # Resultatet uten duplikater

    for row in data:
        date_str = row[0].split(";")[date_column]  # Henter ut dato fra riktig kolonne
        if " " not in date_str:
            date_str += " 00:00:00"  # Legger til standard tid hvis tidspunktet mangler
        if len(date_str) == 16:
            date_str += ":00"  # Legger til standard sekunder hvis tidspunktet mangler
        date_obj = datetime.strptime(date_str, "%d.%m.%Y %H:%M:%S")  # Konverterer til datetime-objekt
        meta = ";".join(row[0].split(";")[2:])  # Henter ut meta-informasjon

        if meta not in last_seen_data:  # Hvis ikke tidligere sett, legg til i resultatet og oppdater siste sett tidspunkt
            last_seen_data[meta] = date_obj
            result.append(row)
        else: # Hvis tidligere sett, sjekk tidsdifferanse og oppdater siste sett tidspunkt hvis nødvendig
            diff = date_obj - last_seen_data[meta]
            diff_hour = diff / timedelta(hours=1)

            if diff_hour > hours_between_alarms:
                # Hvis tidsdifferansen er større enn hours_between_alarms, legg til raden i resultatet og oppdater
                # siste sett dato
                last_seen_data[meta] = date_obj
                result.append(row)
    return result


def create_calendar(start_time_str, end_time_str, days, inside_time_range):
    """
        Oppretter et kalenderdiagram som viser tidspunkt og dager basert på filtreringsparametere.
        :param start_time_str: Starttidspunkt som streng i formatet "HH:MM:SS".
        :param end_time_str: Sluttidspunkt som streng i formatet "HH:MM:SS".
        :param days: Liste over ukedager (0-6) som skal vises.
        :param inside_time_range: Angir om tidspunktene skal være innenfor eller utenfor det angitte tidsintervallet.
        """

    # Opprett figurobjekt og akselobjekt
    fig, ax = plt.subplots(figsize=(10, 6))

    # Sett tittel på aksen og navn på aksene
    ax.set_title('Datofilter', fontsize=14)

    # Definer ukedager og timer
    day_names = ['Mandag', 'Tirsdag', 'Onsdag',
                 'Torsdag', 'Fredag', 'Lørdag',
                 'Søndag']
    hours = [str(i).zfill(2) + ':00' for i in range(25)]
    hours.reverse()

    # Sett y-aksen til å vise timer fra 00:00 til 23:59
    ax.set_yticks(range(0, 25))
    ax.set_yticklabels(hours)

    # Sett x-aksen til å vise ukedager
    ax.set_xlim([0, 7])
    ax.set_xticks([0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5])
    ax.set_xticklabels(day_names, fontsize=10)

    start_time_obj = datetime.strptime(start_time_str, "%H:%M:%S").time()
    end_time_obj = datetime.strptime(end_time_str, "%H:%M:%S").time()

    color = "crimson"
    if inside_time_range:
        color = "green"
        ax.add_patch(plt.Rectangle((0, 0), 7, 24, facecolor="crimson"))
    else:
        ax.add_patch(plt.Rectangle((0, 0), 7, 24, facecolor="green"))
    for day in days:
        if start_time_obj < end_time_obj:
            x = day
            y = 24 - end_time_obj.hour - end_time_obj.minute / 60
            length = end_time_obj.hour - end_time_obj.minute / 60 - start_time_obj.hour - start_time_obj.minute / 60
            ax.add_patch(plt.Rectangle((x, y), 1, length, facecolor=color))
        else:
            x = day
            ax.add_patch(plt.Rectangle((x, 24), 1, -end_time_obj.hour - end_time_obj.minute / 60, facecolor=color))
            ax.add_patch(
                plt.Rectangle((x, 0), 1, 24 - start_time_obj.hour - start_time_obj.minute / 60, facecolor=color))

    # Legg til rektangelobjekter for å representere timer og dager
    for i in range(7):
        for j in range(24):
            ax.add_patch(plt.Rectangle((i, j), 1, 1, facecolor="none", edgecolor='black'))
    plt.show()


def filter_file(filename, auto_label, blacklist, whitelist_any, whitelist_all, date_filter_on, start_date_str,
                end_date_str, start_time_str, end_time_str, days, inside_time_range, show_columns
                , hours_between_alarms, case_sensitive, date_column):
    """
    Leser en CSV-fil og genererer et søylediagram med antall forekomster av hver alarm.

    :param filename: Navnet på CSV-filen som skal leses.
    :param auto_label: Angir om den første linjen i CSV-filen skal brukes som aksetittel automatisk.
    :param blacklist: En liste over ord som skal filtrere ut rader.
    :param whitelist_any: En liste over ord hvor minst én av dem må matche for å inkludere en rad.
    :param whitelist_all: En liste over ord hvor alle må matche for å inkludere en rad.
    :param date_filter_on: Angir om filtrering basert på dato skal aktiveres.
    :param start_date_str: Startdato for filtrering (format: "YYYY-MM-DD").
    :param end_date_str: Sluttdato for filtrering (format: "YYYY-MM-DD").
    :param start_time_str: Starttidspunkt for filtrering (format: "HH:MM:SS").
    :param end_time_str: Sluttidspunkt for filtrering (format: "HH:MM:SS").
    :param days: En liste over ukedager (0-6) som skal inkluderes i filtreringen.
    :param inside_time_range: Angir om tidsintervallet skal være inkluderende (True) eller ekskluderende (False).
    :param show_columns: En liste over kolonner som skal vises i resultatet.
    :param hours_between_alarms: Antall timer som må gå mellom to alarmer for at den andre skal regnes som et nytt tilfelle.
    :param case_sensitive: Angir om filtreringen av teksten skal være følsom for store/små bokstaver (True) eller ikke (False).
    :param date_column: Indeksen til kolonnen i dataene som inneholder tidspunktet for alarmene.
    :return: Etikettene for søylediagrammet og en dictionary med antall forekomster av hver alarm.
    """

    # Åpner CSV-filen
    with open(filename, "r", encoding="cp1250") as file:
        reader = csv.reader(file)

        # Liste for elementer som ikke filtreres bort.
        data = []
        label = "Test"  # Standard aksetittel
        if auto_label:
            label = next(reader)[0].split(";")  # Henter aksetittel fra første linje i CSV-filen

        # Går gjennom hver rad i csv-filen
        for row in reader:

            # Fikser problem med komma i Alarm.Text og Alarm.Value
            new_row = [""]
            if len(row) > 1:
                for r in row:
                    new_row[0] += f"{r},"
                row = new_row

            # Fikser tegnsett
            row[0] = row[0].replace("†", "å").replace("›", "ø").replace("ť", "Ø")

            # Hopper over rader som blir filtrert ut
            if filter_row(row, blacklist, whitelist_any, whitelist_all, case_sensitive):
                continue

            # Hopper over rader som blir filtrert ut basert på dato
            if date_filter_on:
                if not filter_date(row, start_date_str, end_date_str, start_time_str, end_time_str, days,
                                   inside_time_range, date_column):
                    continue

            # Legger til alarmer som skal vises i resultatet
            data.append(row)

        # Fjerner duplikater basert på tid mellom alarmer
        data = filter_duplicates(data, hours_between_alarms, date_column)

        # Teller antall forekomster av like alarmer
        result = {}
        for row in data:
            temp = []
            for i, text in enumerate(row[0].split(";")):
                if i in show_columns:
                    temp.append(text)
            row = ";".join(temp)
            if row in result:
                result[row] += 1
            else:
                result[row] = 1
        return label, result


def draw_bar_chart(data, date_filter, bar_count, show_columns, label, keywords, colors, time_title):
    """
    Lager et stolpediagram basert på dataene.
    :param data: En dictionary som inneholder dataene som skal vises i diagrammet.
    :param date_filter: Angir om det er et datofilter som er brukt.
    :param bar_count: Antall stolper som skal vises i diagrammet.
    :param show_columns: Kolonnene som skal vises.
    :param label: Navn på aksen som viser kategoriene.
    :param keywords: Nøkkelord for fargekoding av stolpene.
    :param colors: Farger for stolpene basert på nøkkelord.
    :param time_title: Tittel for diagrammet relatert til tidspunkt.
    """

    # Sorterer dataen i synkende rekkefølge basert på antall
    sorted_data = dict(sorted(data.items(), key=lambda x: x[1], reverse=True)[:bar_count])

    rows = list(sorted_data.keys())
    count = list(sorted_data.values())

    # Bruker tabulate-pakken for å formaterer radene til like brede kolonner
    cells = tabulate([row.split(";") for row in rows], tablefmt="github").split("|")
    column_count = len(show_columns)+1
    rows = ["".join(cells[i + 1:i + column_count]) for i in range(0, len(cells), column_count)][1:-1]

    # Setter farge på stolpene
    color_list = []
    for row in rows:
        color = "dimgray"
        for i, keyword in enumerate(keywords):
            if isinstance(keyword, str):
                if keyword in row:
                    color = colors[i]
            elif isinstance(keyword, list):
                for sub_keyword in keyword:
                    if sub_keyword in row:
                        color = colors[i]
        color_list.append(color)

    # Lag plottet
    fig, ax = plt.subplots(figsize=(10, 8))
    font_type = "Consolas"
    if platform.system() == "Darwin":
        font_type = "Menlo"
    font = {'fontname': font_type, "fontsize": "12"}

    # Setter posisjon for kategorinavnene
    ax.set_yticks(ticks=range(len(rows)))
    # Setter navn på kategorinavnene
    ax.set_yticklabels(rows, **font)

    for i, row in enumerate(rows):
        # Lager stolpene
        bar = ax.barh(row, count[i], color=color_list[i], edgecolor="black")
        # Setter merkelapp i midten eller på kanten av stolpene basert på stolpens lengde
        if count[i] < count[0] / 2:
            ax.bar_label(bar, label_type='edge', color="black", **font, padding=5)
        else:
            ax.bar_label(bar, label_type='center', color="white", **font)

    # Setter tittel
    if date_filter:
        ax.set_title("Alarmdata. " + time_title)
    else:
        ax.set_title("Alarmdata")

    # Setter tittel for x-aksen
    ax.set_xlabel("Antall alarmer", **font)

    # Setter tittel for y-aksen
    if isinstance(label, list):
        temp = []
        for i, el in enumerate(label):
            if i in show_columns:
                temp.append(el)
        ax.set_ylabel(temp, **font)
    else:
        ax.set_ylabel(label, **font)

    fig.tight_layout()
    plt.savefig("Autosaved.png")
    plt.show()


# Kjører funksjonen "configuration" når denne filen kjøres
if __name__ == "__main__":
    configuration()
